/**
 * TimeoutTask class creates an instance for managing and executing a given asynchronous function with a specified timeout duration.
 * The task will be automatically interrupted if it exceeds the given timeout duration, and will throw a custom error message.
 * Additional information such as execution time can be retrieved using getTime method after the task has been executed.
 *
 * @typeparam T - The return type of the asynchronous function to be executed.
 */
export declare class TimeoutTask<T> {
    private fn;
    private timeout;
    private interruptPromise;
    private interrupt;
    private totalTime;
    constructor(fn: () => Promise<T>, timeout?: number, fnName?: string);
    /**
     * Executes the given function with a specified timeout.
     * If the function takes longer than the timeout, it will be interrupted and an error will be thrown.
     * The total execution time of the function will be stored in the totalTime property.
     *
     * @returns The result of the executed function if completed within the timeout.
     * @throws An error with a message indicating the function was interrupted due to exceeding the specified timeout.
     */
    exec(): Promise<T>;
    /**
     * Returns the interrupt promise associated with the TimeoutTask instance.
     * The interrupt promise is used internally to reject the task when a timeout occurs.
     * This method can be helpful when debugging or tracking the state of the task.
     *
     * @returns The interrupt promise associated with the task.
     */
    getInterruptPromise(): Promise<any>;
    /**
     * Get the total time spent on the most recent execution of the wrapped function.
     * This method provides the duration from the start to the end of the function execution, whether it completed or timed out.
     *
     * @returns The total time in milliseconds spent on the most recent function execution.
     */
    getTime(): number;
}
export declare const executeTimeout: <T>(fn: () => Promise<T>, timeout?: number, fnName?: string) => Promise<T>;
//# sourceMappingURL=timeout.d.ts.map