declare const $transferable: unique symbol;
/**
 * Represents a descriptor for transferable objects in multi-threaded environments.
 * Provides a structure for marking certain objects as transferable and managing the ownership transfer
 * between threads, particularly useful when working with Web Workers.
 */
export interface TransferDescriptor<T = any> {
    /**
     * A unique symbol indicating that an object is a TransferDescriptor.
     */
    [$transferable]: true;
    /**
     * The transferable data to be sent between threads.
     */
    send: T;
    /**
     * An array of objects that can be transferred between threads without serialization and deserialization.
     */
    transferables: Transferable[];
}
/**
 * Determines whether a given object is a TransferDescriptor.
 * A TransferDescriptor is an object with a [$transferable] property set to true and used for
 * transferring ownership of transferable objects between threads.
 * This function checks if the input object has the required properties to be considered
 * a valid TransferDescriptor.
 *
 * @param thing - The object to be checked for being a TransferDescriptor.
 * @returns True if the object is a TransferDescriptor, false otherwise.
 */
export declare function isTransferDescriptor(thing: any): thing is TransferDescriptor;
/**
 * Mark a transferable object as such, so it will no be serialized and
 * deserialized on messaging with the main thread, but to transfer
 * ownership of it to the receiving thread.
 *
 * Only works with array buffers, message ports and few more special
 * types of objects, but it's much faster than serializing and
 * deserializing them.
 *
 * Note:
 * The transferable object cannot be accessed by this thread again
 * unless the receiving thread transfers it back again!
 *
 * @param transferable - Array buffer, message port or similar.
 * @see https://developers.google.com/web/updates/2011/12/Transferable-Objects-Lightning-Fast
 */
export declare function Transfer<T>(transferable: Transferable): TransferDescriptor<T>;
/**
 * Mark transferable objects within an arbitrary object or array as
 * being a transferable object. They will then not be serialized
 * and deserialized on messaging with the main thread, but ownership
 * of them will be transferred to the receiving thread.
 *
 * Only array buffers, message ports and few more special types of
 * objects can be transferred, but it's much faster than serializing and
 * deserializing them.
 *
 * Note:
 * The transferable object cannot be accessed by this thread again
 * unless the receiving thread transfers it back again!
 *
 * @param transferable - Array buffer, message port or similar.
 * @see https://developers.google.com/web/updates/2011/12/Transferable-Objects-Lightning-Fast
 */
export declare function Transfer<T>(payload: T, transferables: Transferable[]): TransferDescriptor<T>;
export {};
//# sourceMappingURL=transferable.d.ts.map