import { Listener } from './interface/listener.js';
/**
 * Keeps track of clients, providing a broadcast, and request/response api with multiplexing.
 */
export declare class TransportServer<Payload> {
    private listener;
    private msgHandlerFn;
    private sockets;
    constructor(listener: Listener, msgHandlerFn: (msg: Payload) => Promise<any>);
    /**
     * Starts the TransportServer, allowing it to accept new connections and handle incoming messages.
     * The server will listen for 'new_socket' events from the underlying listener and invoke the provided message handler function
     * for each received message. The server remains active until the 'stop' method is called.
     */
    start(): void;
    /**
     * Stops accepting new connections. It doesn't close existing sockets.
     * It's expected the clients will gracefully complete by closing their end, sending an `undefined` message.
     */
    stop(): void;
    /**
     * Sends a broadcast message to all connected clients.
     * The given payload will be sent to all the clients currently connected to the TransportServer.
     * It waits for all the messages to be sent and resolves when they are all sent successfully.
     *
     * @param msg - The payload to broadcast to all connected clients.
     * @returns A Promise that resolves when all messages have been sent successfully.
     */
    broadcast(msg: Payload): Promise<void>;
    /**
     * Handles the addition of a new socket to the server by registering a message handler for the client
     * and adding the socket to the list of active sockets. The message handler processes incoming messages
     * from the client, including detecting client disconnection and removing the closed socket.
     *
     * @param socket - The new Socket instance that has connected to the server.
     */
    private handleNewSocket;
    /**
     * Detect the 'transferables' argument to our socket from our message
     * handler return type.
     * @param data - The compound payload data.
     * @returns The split data and transferables.
     */
    private getPayloadAndTransfers;
    /**
     * Handles incoming socket messages, processing the request and sending back a response.
     * This function is responsible for invoking the registered message handler function with the received
     * payload, extracting the result and transferables, and sending a response message back to the client.
     * In case of an error during message handling, it sends an error response with the stack trace.
     *
     * @param socket - The Socket instance from which the message was received.
     * @param msg - The RequestMessage object containing the message ID and payload.
     */
    private handleSocketMessage;
}
//# sourceMappingURL=transport_server.d.ts.map